@@ -1,7 +1,7 @@
 ;;; xcb-keysyms.el --- Conversion between  -*- lexical-binding: t -*-
 ;;;                    X keysyms, X keycodes and Emacs key event.
 
-;; Copyright (C) 2015-2019 Free Software Foundation, Inc.
+;; Copyright (C) 2015-2025 Free Software Foundation, Inc.
 
 ;; Author: Chris Feng <chris.w.feng@gmail.com>
 
@@ -18,7 +18,7 @@
 ;; GNU General Public License for more details.
 
 ;; You should have received a copy of the GNU General Public License
-;; along with GNU Emacs.  If not, see <http://www.gnu.org/licenses/>.
+;; along with GNU Emacs.  If not, see <https://www.gnu.org/licenses/>.
 
 ;;; Commentary:
 
@@ -33,13 +33,13 @@
 ;;   'xcb:KeyButMask:Mod*'.
 
 ;; References:
-;; + X protocol (http://www.x.org/releases/X11R7.7/doc/xproto/x11protocol.txt)
+;; + X protocol (https://www.x.org/releases/X11R7.7/doc/xproto/x11protocol.txt)
 ;; + XKB protocol (https://www.x.org/releases/X11R7.7/doc/kbproto/xkbproto.txt)
 ;; + xcb/util-keysyms (git://anongit.freedesktop.org/xcb/util-keysyms)
 
 ;;; Code:
 
-(eval-when-compile (require 'cl-lib))
+(require 'cl-lib)
 
 (require 'xcb)
 (require 'xcb-xkb)
@@ -103,11 +103,11 @@ This method must be called before using any other method in this module."
     (xcb:keysyms:-update-modkeys obj xcb:xkb:ID:UseCoreKbd)
     ;; Attach event listeners.
     (xcb:+event obj 'xcb:xkb:NewKeyboardNotify
-                `(lambda (data _)
-                   (xcb:keysyms:-on-NewKeyboardNotify ,obj data)))
+                (lambda (data _)
+                  (xcb:keysyms:-on-NewKeyboardNotify obj data)))
     (xcb:+event obj 'xcb:xkb:MapNotify
-                `(lambda (data _)
-                   (xcb:keysyms:-on-MapNotify ,obj data)))
+                (lambda (data _)
+                  (xcb:keysyms:-on-MapNotify obj data)))
     ;; Select XKB MapNotify and NewKeyboardNotify events.
     (let ((map (logior xcb:xkb:MapPart:KeyTypes
                        xcb:xkb:MapPart:KeySyms
@@ -149,7 +149,7 @@ This method must be called before using any other method in this module."
                        :autoCtrlsValues 0))))
 
 (cl-defmethod xcb:keysyms:-on-NewKeyboardNotify ((obj xcb:connection) data)
-  "Handle 'NewKeyboardNotify' event."
+  "Handle a \\='NewKeyboardNotify' event."
   (let ((device-id (xcb:-get-extra-plist obj 'keysyms 'device-id))
         (callback (xcb:-get-extra-plist obj 'keysyms 'callback))
         (obj1 (make-instance 'xcb:xkb:NewKeyboardNotify))
@@ -182,7 +182,7 @@ This method must be called before using any other method in this module."
         (funcall callback)))))
 
 (cl-defmethod xcb:keysyms:-on-MapNotify ((obj xcb:connection) data)
-  "Handle 'MapNotify' event."
+  "Handle \\='MapNotify' event."
   (let ((device-id (xcb:-get-extra-plist obj 'keysyms 'device-id))
         (callback (xcb:-get-extra-plist obj 'keysyms 'callback))
         (obj1 (make-instance 'xcb:xkb:MapNotify))
@@ -322,7 +322,7 @@ FIRST-KEYCODE and COUNT specify the keycode range to update."
           xcb:keysyms:shift-lock-mask 0
           xcb:keysyms:num-lock-mask 0)
     (dolist (row (number-sequence 3 7))
-      (let ((mask (lsh 1 row))
+      (let ((mask (ash 1 row))
             (col 0)
             found-alt-or-meta keycode keysym)
         (while (< col keycodes-per-modifier)
@@ -371,9 +371,14 @@ FIRST-KEYCODE and COUNT specify the keycode range to update."
 
 (cl-defmethod xcb:keysyms:keycode->keysym ((obj xcb:connection) keycode
                                            modifiers)
-  "Convert keycode to (keysym . mod-mask).
+  "Convert KEYCODE to keysym or get possible modifier combinations for keycode.
 
-Return (0 . 0) when conversion fails."
+If MODIFIERS is non-nil, return (KEYSYM . CONSUMED-MODIFIERS) where
+CONSUMED-MODIFIERS should be lognot'ed with MODIFIERS so as to make further
+conversion correct.  (0 . 0) is returned when conversion fails.
+
+If MODIFIERS is nil, return all possible modifier combinations for this
+keycode.  The caller is responsible for checking which modifiers to use."
   (let ((preserve 0)
         group group-info group-number index keytype)
     (with-slots (keytypes keycodes min-keycode max-keycode)
@@ -390,12 +395,14 @@ Return (0 . 0) when conversion fails."
         ;; Check group number.
         (when (= group-number 0)
           (throw 'return '(0 . 0)))
-        (setq group (logand (lsh modifiers -13) #b11)) ;The 13, 14 bits.
+        (setq group (if (null modifiers)
+                        0
+                      (logand (ash modifiers -13) #b11))) ;The 13, 14 bits.
         ;; Wrap group.
         (when (>= group group-number)
           (pcase (logand group-info #xC0) ;See <XKBstr.h>.
             (`xcb:xkb:GroupsWrap:RedirectIntoRange
-             (setq group (logand #xFF (lsh group-info -4))) ;See <XKBstr.h>.
+             (setq group (logand #xFF (ash group-info -4))) ;See <XKBstr.h>.
              ;; Check if it's also out of range.
              (when (>= group group-number)
                (setq group 0)))
@@ -408,23 +415,28 @@ Return (0 . 0) when conversion fails."
         ;; Get key type.
         (setq keytype (aref keytypes
                             (elt (slot-value keycode 'kt-index) group)))
-        ;; Find the shift level and preserved modifiers.
         (with-slots (mods-mask hasPreserve map (preserve* preserve)) keytype
-          (catch 'break
-            (dolist (entry map)
-              (with-slots (active (mods-mask* mods-mask) level) entry
-                (when (and (= 1 active)
-                           (= (logand modifiers mods-mask) mods-mask*))
-                  (cl-incf index level)
-                  (when (= 1 hasPreserve)
-                    (setq preserve (slot-value (elt preserve*
-                                                    (cl-position entry map))
-                                               'mask)))
-                  (throw 'break nil)))))
-          ;; FIXME: Use of preserved modifiers (e.g. capitalize the keysym
-          ;;        when LOCK is preserved)?
-          (cons (elt (slot-value keycode 'syms) index)
-                (logand mods-mask (lognot preserve))))))))
+          (if (null modifiers)
+              ;; Return all possible modifier combinations.
+              (delq nil
+                    (mapcar (lambda (entry)
+                              (when (= (slot-value entry 'active) 1)
+                                (slot-value entry 'mods-mask)))
+                            map))
+            ;; Find the shift level and preserved modifiers.
+            (catch 'break
+              (dolist (entry map)
+                (with-slots (active (mods-mask* mods-mask) level) entry
+                  (when (and (= 1 active)
+                             (= (logand modifiers mods-mask) mods-mask*))
+                    (cl-incf index level)
+                    (when (= 1 hasPreserve)
+                      (setq preserve (slot-value (elt preserve*
+                                                      (cl-position entry map))
+                                                 'mask)))
+                    (throw 'break nil)))))
+            (cons (elt (slot-value keycode 'syms) index)
+                  (logand mods-mask (lognot preserve)))))))))
 
 (cl-defmethod xcb:keysyms:keysym->keycode ((obj xcb:connection) keysym)
   "Convert keysym to (the first matching) keycode.
@@ -573,10 +585,10 @@ Return 0 if conversion fails."
         ,@(make-list 64 nil)]
   "Emacs event representations of XF86keysym (#x1008ff00 - #x1008ffff).")
 
-(cl-defmethod xcb:keysyms:event->keysym ((obj xcb:connection) event)
-  "Translate Emacs key event EVENT to (keysym . mod-mask).
+(cl-defmethod xcb:keysyms:event->keysyms ((obj xcb:connection) event)
+  "Translate Emacs key event EVENT to list of (keysym . mod-mask).
 
-Return (0 . 0) when conversion fails."
+Return ((0 . 0)) when conversion fails."
   (let ((modifiers (event-modifiers event))
         (event (event-basic-type event))
         keysym)
@@ -626,14 +638,7 @@ Return (0 . 0) when conversion fails."
                                 (throw 'break key)))
                             x-keysym-table)))))))
     (if (not keysym)
-        '(0 . 0)
-      (let ((keycode (xcb:keysyms:keysym->keycode obj keysym))
-            keysym*)
-        (when (/= 0 keycode)
-          (setq keysym* (xcb:keysyms:keycode->keysym obj keycode 0))
-          (unless (= keysym (car keysym*))
-            ;; This keysym requires additional modifiers to input.
-            (push (cdr keysym*) modifiers))))
+        '((0 . 0))
       (when modifiers
         ;; Do transforms: * -> x-*-keysym -> xcb:keysyms:*-mask.
         (setq modifiers (mapcar (lambda (i)
@@ -674,8 +679,26 @@ Return (0 . 0) when conversion fails."
                                     (_
                                      ;; Include but not limit to: down.
                                      0)))
-                                modifiers)))
-      (cons keysym (apply #'logior modifiers)))))
+                                modifiers)
+              modifiers (apply #'logior modifiers)))
+      (let ((keycode (xcb:keysyms:keysym->keycode obj keysym))
+            extra-modifiers)
+        (when (/= 0 keycode)
+          (setq extra-modifiers (xcb:keysyms:keycode->keysym obj keycode nil)
+                ;; Always try without other modifier.
+                extra-modifiers (append '(0) extra-modifiers)
+                ;; Keep all modifiers helping convert keycode to this keysym.
+                extra-modifiers
+                (delq nil
+                      (mapcar (lambda (modifier)
+                                (when (= (car (xcb:keysyms:keycode->keysym
+                                               obj keycode modifier))
+                                         keysym)
+                                  modifier))
+                              extra-modifiers))))
+        (mapcar (lambda (extra-modifier)
+                  (cons keysym (logior (or modifiers 0) extra-modifier)))
+                extra-modifiers)))))
 
 (cl-defmethod xcb:keysyms:keysym->event ((_obj xcb:connection) keysym
                                          &optional mask allow-modifiers)
