@@ -1,6 +1,6 @@
 ;;; xcb.el --- X protocol Emacs Lisp Binding (XELB)  -*- lexical-binding: t -*-
 
-;; Copyright (C) 2015-2019 Free Software Foundation, Inc.
+;; Copyright (C) 2015-2025 Free Software Foundation, Inc.
 
 ;; Author: Chris Feng <chris.w.feng@gmail.com>
 
@@ -17,7 +17,7 @@
 ;; GNU General Public License for more details.
 
 ;; You should have received a copy of the GNU General Public License
-;; along with GNU Emacs.  If not, see <http://www.gnu.org/licenses/>.
+;; along with GNU Emacs.  If not, see <https://www.gnu.org/licenses/>.
 
 ;;; Commentary:
 
@@ -48,11 +48,12 @@
 ;; + Use XC-MISC extension for `xcb:generate-id' when IDs are used up.
 
 ;; References:
-;; + X protocol (http://www.x.org/releases/X11R7.7/doc/xproto/x11protocol.txt)
+;; + X protocol (https://www.x.org/releases/X11R7.7/doc/xproto/x11protocol.txt)
 
 ;;; Code:
 
-(eval-when-compile (require 'cl-lib))
+(require 'cl-lib)
+
 (require 'xcb-xproto)
 
 (defvar xcb:connection-timeout 3 "Connection timeout.")
@@ -186,7 +187,8 @@
                                                 socket)))
       (setq display (frame-parameter nil 'display)
             socket (xcb:display->socket display)))
-    (let* ((process (make-network-process :name "XELB" :remote socket))
+    (let* ((process (make-network-process :name "XELB" :family 'local
+                                          :service socket))
            (auth (if auth-info auth-info (xcb:create-auth-info)))
            (connection (make-instance 'xcb:connection
                                       :process process :display display
@@ -204,6 +206,7 @@
                        (plist-put (process-plist process) 'connection obj))
     (set-process-coding-system process 'binary 'binary)
     (set-process-filter process #'xcb:-connection-setup-filter)
+    (set-process-sentinel process #'xcb:-connection-sentinel)
     (process-send-string                ;send setup packet
      process
      (apply #'unibyte-string
@@ -267,6 +270,11 @@
             (x (error "Unrecognized setup status: %d" x)))))
       (setf (slot-value connection 'lock) nil))))
 
+(defun xcb:-connection-sentinel (process _event)
+  "Process sentinel used to teardown the connection on disconnect."
+  (unless (process-live-p process)
+    (xcb:disconnect (plist-get (process-plist process) 'connection))))
+
 (cl-defmethod xcb:-convert-sequence ((obj xcb:connection) sequence16)
   "Convert 16-bit sequence number SEQUENCE16 (read from a packet).
 
@@ -289,6 +297,9 @@ The result would be 29 or 61 bits, depending on the machine."
 
 Concurrency is disabled as it breaks the orders of errors, replies and events."
   (let* ((connection (plist-get (process-plist process) 'connection))
+         ;; Temporarily disable GC here as typically it's about to do
+         ;; lots of marshaling/unmarshaling.
+         (gc-cons-threshold most-positive-fixnum)
          (cache (vconcat (slot-value connection 'message-cache) message))
          (cache-length (length cache)))
     (setf (slot-value connection 'message-cache) cache)
@@ -364,17 +375,21 @@ Concurrency is disabled as it breaks the orders of errors, replies and events."
                     ;; Too short.
                     (throw 'break nil))
                   (setq listener
-                        (lax-plist-get (slot-value connection 'event-plist)
-                                       (vector (aref cache 1)
-                                               (funcall
-                                                (if xcb:lsb
-                                                    #'xcb:-unpack-u2-lsb
-                                                  #'xcb:-unpack-u2)
-                                                cache 8)))))
+                        (compat-call plist-get
+                                     (slot-value connection 'event-plist)
+                                     (vector (aref cache 1)
+                                             (funcall
+                                              (if xcb:lsb
+                                                  #'xcb:-unpack-u2-lsb
+                                                #'xcb:-unpack-u2)
+                                              cache 8))
+                                     #'equal)))
                  (`xkb
                   (setq listener
-                        (lax-plist-get (slot-value connection 'event-plist)
-                                       (vector (aref cache 1))))))
+                        (compat-call plist-get
+                                     (slot-value connection 'event-plist)
+                                     (vector (aref cache 1))
+                                     #'equal))))
                ;; Conventional events are 32 bytes in size.
                (unless event-length
                  (setq event-length 32))
@@ -408,15 +423,14 @@ Concurrency is disabled as it breaks the orders of errors, replies and events."
               (setq data (aref event 1)
                     synthetic (aref event 2))
               (dolist (listener (aref event 0))
-                (unwind-protect
-                    (xcb-debug:backtrace-on-error
-                     (funcall listener data synthetic))))))
+                (xcb-debug:backtrace-on-error
+                 (funcall listener data synthetic)))))
         (cl-decf event-lock)))))
 
 (cl-defmethod xcb:disconnect ((obj xcb:connection))
   "Disconnect from X server."
   (when (slot-value obj 'connected)
-    (xcb:flush obj)
+    (ignore-errors (xcb:flush obj))
     (delete-process (slot-value obj 'process))
     ;; Reset every slot to its default value
     (let ((slots (eieio-class-slots 'xcb:connection)))
@@ -455,10 +469,10 @@ classes of EVENT (since they have the same event number)."
             (plist-put plist key
                        (if (child-of-class-p event 'xcb:-generic-event)
                            'xge 'xkb))))
-    (setq listeners (lax-plist-get plist event-number))
+    (setq listeners (compat-call plist-get plist event-number #'equal))
     (setf (slot-value obj 'event-plist)
-          (lax-plist-put plist event-number (append listeners
-                                                    (list listener))))))
+          (compat-call plist-put plist event-number
+                       (append listeners (list listener)) #'equal))))
 
 (cl-defmethod xcb:flush ((obj xcb:connection))
   "Flush request data to X server."
@@ -542,10 +556,10 @@ classes of EVENT (since they have the same event number)."
          (cache (slot-value obj 'request-cache)))
     (when extension-opcode
       (setq msg (vconcat (vector extension-opcode) msg))
-      (setq len (1+ len)))
+      (cl-incf len))
     (when (> 2 (length msg))   ;for short message (e.g. GetInputFocus)
       (setq msg (vconcat msg [0]))
-      (setq len (1+ len)))
+      (cl-incf len))
     (setq msg
           (vconcat (substring msg 0 2)
                    (funcall (if (slot-value request '~lsb) #'xcb:-pack-u2-lsb
